package ethereum_test

import (
	"context"
	"errors"
	"math/big"
	"testing"

	"github.com/ethereum/go-ethereum/common/hexutil"
	"github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/params"
	"github.com/golang/mock/gomock"
	"github.com/mailchain/mailchain/cmd/indexer/internal/actions"
	"github.com/mailchain/mailchain/cmd/indexer/internal/ethereum"
	"github.com/mailchain/mailchain/cmd/internal/datastore"
	"github.com/mailchain/mailchain/cmd/internal/datastore/datastoretest"
	"github.com/mailchain/mailchain/encoding/encodingtest"
	"github.com/stretchr/testify/assert"
)

func TestTransaction_ToTransaction(t *testing.T) {
	mockCtrl := gomock.NewController(t)
	defer mockCtrl.Finish()
	type fields struct {
		txStore    datastore.TransactionStore
		rawTxStore datastore.RawTransactionStore
		pkStore    datastore.PublicKeyStore
		networkID  *params.ChainConfig
	}
	type args struct {
		blk *types.Block
		tx  *types.Transaction
	}
	tests := []struct {
		name    string
		fields  fields
		args    args
		want    *datastore.Transaction
		wantErr bool
	}{
		{
			"success",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c")}, nil, nil),
				getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c"),
			},
			&datastore.Transaction{
				From:      []uint8{0x11, 0x90, 0x58, 0xdc, 0x2c, 0x57, 0x7e, 0x9c, 0x4b, 0xa6, 0x91, 0x46, 0x78, 0xaa, 0x9d, 0xb5, 0x65, 0x30, 0xf, 0xfe},
				To:        []uint8{0xec, 0xa5, 0x6d, 0x4, 0x54, 0x6a, 0xff, 0xce, 0xc0, 0xb3, 0xce, 0x61, 0x97, 0x11, 0x36, 0xf4, 0x97, 0x86, 0x6a, 0x3b},
				Data:      []uint8{0xb6, 0x1d, 0x27, 0xf6, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x8a, 0x6, 0x6f, 0x5f, 0xde, 0xbe, 0x40, 0x9f, 0xc1, 0x93, 0x9f, 0x51, 0x5e, 0x19, 0xa2, 0x81, 0x63, 0x88, 0x91, 0xfb, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xa, 0x60, 0x4b, 0x9a, 0x42, 0xdf, 0x9c, 0xa0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0},
				BlockHash: []uint8{0x52, 0xa7, 0x3a, 0x5c, 0x6d, 0xfc, 0xea, 0x77, 0xa0, 0x32, 0x85, 0x68, 0xad, 0x26, 0x7d, 0xdb, 0xfe, 0xcb, 0x2e, 0xcb, 0x75, 0xa2, 0xfe, 0x5e, 0xed, 0xad, 0xec, 0x58, 0xdb, 0xc, 0xd, 0x93},
				Hash:      []uint8{0xcd, 0x3c, 0xcc, 0x84, 0x6c, 0x56, 0x6f, 0xbf, 0x76, 0xf3, 0x8b, 0x11, 0x84, 0xba, 0x2, 0x26, 0x1a, 0x82, 0x1c, 0x69, 0x42, 0xd1, 0x81, 0x46, 0xb8, 0x9b, 0x9d, 0x28, 0x5a, 0xa2, 0x9b, 0x9c},
				Value:     *big.NewInt(0),
				GasUsed:   *big.NewInt(147670),
				GasPrice:  *big.NewInt(50000000000),
			},
			false,
		},
		{
			"success-to-nil",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "0x157cdd4029ce9bbf319b9d7cc27fd9989e9be63d0427a193f2297e2dee7b4a2e")}, nil, nil),
				getTx(t, "0x157cdd4029ce9bbf319b9d7cc27fd9989e9be63d0427a193f2297e2dee7b4a2e"),
			},
			&datastore.Transaction{
				From: []uint8{0x11, 0x90, 0x58, 0xdc, 0x2c, 0x57, 0x7e, 0x9c, 0x4b, 0xa6, 0x91, 0x46, 0x78, 0xaa, 0x9d, 0xb5, 0x65, 0x30, 0xf, 0xfe},
				To:   []uint8(nil),
				Data: []uint8{0x60, 0x60, 0x60, 0x40, 0x52, 0x60, 0x2, 0x61, 0x1, 0x8, 0x60, 0x0, 0x50, 0x55, 0x60, 0x40, 0x51, 0x61, 0x1, 0x56, 0x38, 0x3, 0x80, 0x61, 0x1, 0x56, 0x83, 0x39, 0x81, 0x1, 0x60, 0x40, 0x52, 0x80, 0x51, 0x60, 0x80, 0x51, 0x60, 0xa0, 0x51, 0x91, 0x90, 0x92, 0x1, 0x91, 0x90, 0x80, 0x83, 0x83, 0x81, 0x51, 0x60, 0x1, 0x90, 0x81, 0x1, 0x81, 0x55, 0x60, 0x0, 0x90, 0x60, 0x1, 0x60, 0xa0, 0x60, 0x2, 0xa, 0x3, 0x32, 0x16, 0x90, 0x60, 0x2, 0x90, 0x60, 0x3, 0x83, 0x90, 0x55, 0x91, 0x83, 0x52, 0x50, 0x61, 0x1, 0x2, 0x60, 0x20, 0x52, 0x60, 0x40, 0x82, 0x20, 0x55, 0x5b, 0x82, 0x51, 0x81, 0x10, 0x15, 0x61, 0x0, 0xeb, 0x57, 0x82, 0x81, 0x81, 0x51, 0x81, 0x10, 0x15, 0x61, 0x0, 0x2, 0x57, 0x90, 0x60, 0x20, 0x1, 0x90, 0x60, 0x20, 0x2, 0x1, 0x51, 0x60, 0x1, 0x60, 0xa0, 0x60, 0x2, 0xa, 0x3, 0x16, 0x60, 0x2, 0x60, 0x0, 0x50, 0x82, 0x60, 0x2, 0x1, 0x61, 0x1, 0x0, 0x81, 0x10, 0x15, 0x61, 0x0, 0x2, 0x57, 0x90, 0x90, 0x1, 0x60, 0x0, 0x50, 0x81, 0x90, 0x55, 0x50, 0x80, 0x60, 0x2, 0x1, 0x61, 0x1, 0x2, 0x60, 0x0, 0x50, 0x60, 0x0, 0x85, 0x84, 0x81, 0x51, 0x81, 0x10, 0x15, 0x61, 0x0, 0x2, 0x57, 0x90, 0x60, 0x20, 0x1, 0x90, 0x60, 0x20, 0x2, 0x1, 0x51, 0x60, 0x1, 0x60, 0xa0, 0x60, 0x2, 0xa, 0x3, 0x16, 0x81, 0x52, 0x60, 0x20, 0x1, 0x90, 0x81, 0x52, 0x60, 0x20, 0x1, 0x60, 0x0, 0x20, 0x60, 0x0, 0x50, 0x81, 0x90, 0x55, 0x50, 0x60, 0x1, 0x1, 0x61, 0x0, 0x60, 0x56, 0x5b, 0x81, 0x60, 0x0, 0x60, 0x0, 0x50, 0x81, 0x90, 0x55, 0x50, 0x50, 0x50, 0x50, 0x80, 0x61, 0x1, 0x5, 0x60, 0x0, 0x50, 0x81, 0x90, 0x55, 0x50, 0x61, 0x1, 0xf, 0x62, 0x1, 0x51, 0x80, 0x42, 0x4, 0x90, 0x56, 0x5b, 0x61, 0x1, 0x7, 0x55, 0x50, 0x50, 0x50, 0x50, 0x60, 0x31, 0x80, 0x61, 0x1, 0x25, 0x60, 0x0, 0x39, 0x60, 0x0, 0xf3, 0x0, 0x36, 0x60, 0x0, 0x80, 0x37, 0x60, 0x20, 0x60, 0x0, 0x36, 0x60, 0x0, 0x34, 0x73, 0x27, 0x39, 0x30, 0xd2, 0x1e, 0x1, 0xee, 0x25, 0xe4, 0xc2, 0x19, 0xb6, 0x32, 0x59, 0xd2, 0x14, 0x87, 0x22, 0x20, 0xa2, 0x61, 0x23, 0x5a, 0x5a, 0x3, 0xf2, 0x15, 0x60, 0x1, 0x57, 0x60, 0x20, 0x60, 0x0, 0xf3, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x60, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x52, 0xb7, 0xd2, 0xdc, 0xc8, 0xc, 0xd2, 0xe4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x11, 0x90, 0x58, 0xdc, 0x2c, 0x57, 0x7e, 0x9c, 0x4b, 0xa6, 0x91, 0x46, 0x78, 0xaa, 0x9d, 0xb5, 0x65, 0x30, 0xf, 0xfe}, BlockHash: []uint8{0x8e, 0x3e, 0x75, 0xc3, 0xe3, 0xd, 0x83, 0x6c, 0x3a, 0xbf, 0xf1, 0x24, 0x2b, 0xd6, 0x33, 0xf, 0xed, 0x3b, 0x8f, 0x1f, 0xc3, 0x57, 0x4, 0xfa, 0xcb, 0x8b, 0xfe, 0xff, 0x96, 0xcf, 0xbf, 0x23}, Hash: []uint8{0x15, 0x7c, 0xdd, 0x40, 0x29, 0xce, 0x9b, 0xbf, 0x31, 0x9b, 0x9d, 0x7c, 0xc2, 0x7f, 0xd9, 0x98, 0x9e, 0x9b, 0xe6, 0x3d, 0x4, 0x27, 0xa1, 0x93, 0xf2, 0x29, 0x7e, 0x2d, 0xee, 0x7b, 0x4a, 0x2e},
				Value:    *big.NewInt(0),
				GasUsed:  *big.NewInt(1000000),
				GasPrice: *big.NewInt(50000000000),
			},
			false,
		},
		{
			"err-tx-not-in-block",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				types.NewBlock(&types.Header{}, []*types.Transaction{}, nil, nil),
				getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c"),
			},
			nil,
			true,
		},
		{
			"err-invalid-from",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "invalid-from")}, nil, nil),
				getTx(t, "invalid-from"),
			},
			nil,
			true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			transaction := ethereum.NewTransactionProcessor(
				tt.fields.txStore,
				tt.fields.rawTxStore,
				tt.fields.pkStore,
				tt.fields.networkID,
			)

			got, err := transaction.ToTransaction(tt.args.blk, tt.args.tx)
			if (err != nil) != tt.wantErr {
				t.Errorf("Transaction.toTransaction() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !assert.Equal(t, tt.want, got) {
				t.Errorf("Transaction.toTransaction() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestTransaction_From(t *testing.T) {
	type fields struct {
		txStore    datastore.TransactionStore
		rawTxStore datastore.RawTransactionStore
		pkStore    datastore.PublicKeyStore
		networkID  *params.ChainConfig
	}
	type args struct {
		blockNo *big.Int
		tx      *types.Transaction
	}
	tests := []struct {
		name    string
		fields  fields
		args    args
		want    []byte
		wantErr bool
	}{
		{
			"success-0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c",
			fields{
				nil,
				nil,
				nil,
				params.MainnetChainConfig,
			},
			args{
				big.NewInt(500000),
				getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c"),
			},
			[]byte{0x11, 0x90, 0x58, 0xdc, 0x2c, 0x57, 0x7e, 0x9c, 0x4b, 0xa6, 0x91, 0x46, 0x78, 0xaa, 0x9d, 0xb5, 0x65, 0x30, 0xf, 0xfe},
			false,
		},
		{
			"success-0x9220257407f78ad91f340f856fe147751a95257783c0c2c288a129d356ab05e4",
			fields{
				nil,
				nil,
				nil,
				params.MainnetChainConfig,
			},
			args{
				hexutil.MustDecodeBig("0x9027d9"),
				getTx(t, "0x9220257407f78ad91f340f856fe147751a95257783c0c2c288a129d356ab05e4"),
			},
			[]byte{0x4c, 0x54, 0x99, 0x90, 0xa7, 0xef, 0x3f, 0xea, 0x87, 0x84, 0x40, 0x6c, 0x1e, 0xec, 0xc9, 0x8b, 0xf4, 0x21, 0x1f, 0xa5},
			false,
		},
		{
			"err-invalid",
			fields{
				nil,
				nil,
				nil,
				params.MainnetChainConfig,
			},
			args{
				big.NewInt(500000),
				getTx(t, "invalid-from"),
			},
			[]byte(nil),
			true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			transaction := ethereum.NewTransactionProcessor(
				tt.fields.txStore,
				tt.fields.rawTxStore,
				tt.fields.pkStore,
				tt.fields.networkID,
			)
			got, err := transaction.From(tt.args.blockNo, tt.args.tx)
			if (err != nil) != tt.wantErr {
				t.Errorf("Transaction.From() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !assert.Equal(t, tt.want, got) {
				t.Errorf("Transaction.From() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestTransaction_Run(t *testing.T) {
	mockCtrl := gomock.NewController(t)
	defer mockCtrl.Finish()
	type fields struct {
		txStore    datastore.TransactionStore
		rawTxStore datastore.RawTransactionStore
		pkStore    datastore.PublicKeyStore
		networkID  *params.ChainConfig
	}
	type args struct {
		ctx      context.Context
		protocol string
		network  string
		tx       interface{}
		txOpts   actions.TransactionOptions
	}
	tests := []struct {
		name    string
		fields  fields
		args    args
		wantErr bool
	}{
		{
			"success",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					m.EXPECT().PutPublicKey(context.Background(), "ethereum", "mainnet", encodingtest.MustDecodeHexZeroX("0x119058dc2c577e9c4ba6914678aa9db565300ffe"), gomock.Any()).Return(nil)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				context.Background(),
				"ethereum",
				"mainnet",
				getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c"),
				&ethereum.TxOptions{types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c")}, nil, nil)},
			},
			false,
		},
		{
			"success-nil-to-transation",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					m.EXPECT().PutPublicKey(context.Background(), "ethereum", "mainnet", encodingtest.MustDecodeHexZeroX("0x119058dc2c577e9c4ba6914678aa9db565300ffe"), gomock.Any()).Return(nil)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				context.Background(),
				"ethereum",
				"mainnet",
				getTx(t, "0x157cdd4029ce9bbf319b9d7cc27fd9989e9be63d0427a193f2297e2dee7b4a2e"),
				&ethereum.TxOptions{types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "0x157cdd4029ce9bbf319b9d7cc27fd9989e9be63d0427a193f2297e2dee7b4a2e")}, nil, nil)},
			},
			false,
		},
		{
			"err-put-public-key",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					m.EXPECT().PutPublicKey(context.Background(), "ethereum", "mainnet", encodingtest.MustDecodeHexZeroX("0x119058dc2c577e9c4ba6914678aa9db565300ffe"), gomock.Any()).Return(errors.New("error"))
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				context.Background(),
				"ethereum",
				"mainnet",
				getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c"),
				&ethereum.TxOptions{types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c")}, nil, nil)},
			},
			true,
		},
		{
			"err-get-public-key",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				context.Background(),
				"ethereum",
				"mainnet",
				getTx(t, "invalid-from"),
				&ethereum.TxOptions{types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "invalid-from")}, nil, nil)},
			},
			true,
		},
		{
			"err-tx-opts",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				context.Background(),
				"ethereum",
				"mainnet",
				getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c"),
				0,
			},
			true,
		},
		{
			"err-tx",
			fields{
				func() datastore.TransactionStore {
					m := datastoretest.NewMockTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.RawTransactionStore {
					m := datastoretest.NewMockRawTransactionStore(mockCtrl)
					return m
				}(),
				func() datastore.PublicKeyStore {
					m := datastoretest.NewMockPublicKeyStore(mockCtrl)
					return m
				}(),
				params.MainnetChainConfig,
			},
			args{
				context.Background(),
				"ethereum",
				"mainnet",
				0,
				&ethereum.TxOptions{types.NewBlock(&types.Header{}, []*types.Transaction{getTx(t, "0xcd3ccc846c566fbf76f38b1184ba02261a821c6942d18146b89b9d285aa29b9c")}, nil, nil)},
			},
			true,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			transaction := ethereum.NewTransactionProcessor(
				tt.fields.txStore,
				tt.fields.rawTxStore,
				tt.fields.pkStore,
				tt.fields.networkID,
			)
			if err := transaction.Run(tt.args.ctx, tt.args.protocol, tt.args.network, tt.args.tx, tt.args.txOpts); (err != nil) != tt.wantErr {
				t.Errorf("Transaction.Run() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}
